# encoding: utf-8
__copyright__ = "Copyright (C) 2015 University of Wisconsin SSEC. All rights reserved."

import io
import sys
import logging
from collections import deque

from edosl0util.stream import (
    jpss_packet_stream,
    NonConsecutiveSeqId,
    PacketTooShort
)

LOG = logging.getLogger(__name__)


class InvalidPacketGroup(Exception):
    """
    If a there exists packets in a group with different apids or if the group
    does not end in a last packet.
    """


def _group_packets(stream):
    """
    Returns a generator that yields all packets until a timestamps, i.e., a
    packet group.
    """
    packet = stream.next()
    while not packet.stamp:
        yield packet
        packet = stream.next()
    # put the next packet with a stamp so it will be next
    stream.push_back(packet)


class StreamManager(object):

    def __init__(self, streams):
        self.streams = list(streams)
        self.current_idx = 0
        self.valid = [True]*len(self.streams)
        self.tried = [False]*len(self.streams)
        self.handling_error = False

    def invalid(self, stream):
        idx = self.streams.index(stream)
        self.valid[idx] = False
        self.current_idx = 0

    end_of_stream = invalid

    def have_valid(self):
        return any(self.valid)

    def set_error(self, stream):
        idx = self.streams.index(stream)
        self.tried[idx] = True

    missing_seqid = set_error
    invalid_group = set_error

    def next(self):
        if self.handling_error:
            if all(self.tried):
                # tried every stream, can't recover, move on
                self.tried = [False]*len(self.stream)
                self.handling_error = False
            else:
                # there is a stream we haven't tried, use it
                self.current_idx = self.tried.index(False)
                self.tried[self.current_idx] = True
                return self.streams[self.current_idx]

        self.current_idx = self.valid.index(True, self.current_idx)
        return self.streams[self.current_idx]


def next_stamp_and_apid(stream, stamp, apid):
    LOG.debug("seeking to %s apid %s", stamp, apid)
    stream.seek_to(stamp, apid)
    stream.seek_to_next_stamp()


def next_stamp(stream, stamp):
    LOG.debug("seeking to %s", stamp)
    stream.seek_to(stamp)


def merge(streams, output=sys.stdout):
    last_packet = None

    streams = StreamManager(streams)
    stream = streams.next()
    while streams.have_valid():
        LOG.debug("stream %s", stream)
        LOG.debug("streams valid: %s", streams.valid)
        LOG.debug("streams handling_error:%s tried:%s", streams.handling_error, streams.tried)
        LOG.debug("streams current_idx: %s", streams.current_idx)
        try:
            # Do until `next` causes StopIteration
            while True:
                packets_to_write = deque()
                packet = stream.next()
                if packet.is_standalone():
                    packets_to_write.append(packet)

                elif packet.is_first():  # packet group
                    group = deque([packet])
                    group.extend(_group_packets(stream))
                    if not group[0].is_first() \
                            and group[-1].is_last() \
                            and group[0].apid == group[-1].apid:
                        raise InvalidPacketGroup()
                    packets_to_write.extend(group)

                else:
                    LOG.debug("skipping hanging packet: %s", packet)

                # First packet always has a stamp because it's either
                # standalone or part of a valid group
                last_packet = packets_to_write[0]
                while packets_to_write:
                    pkt = packets_to_write.popleft()
                    output.write(pkt.bytes())

        except NonConsecutiveSeqId:
            LOG.debug('missing sequence id, next stream')
            streams.missing_seqid(stream)

            LOG.debug("seeking to %s apid %s", last_packet.stamp, last_packet.apid)
            stream = streams.next()

        except InvalidPacketGroup:
            LOG.debug("invalid group, next stream")
            streams.invalid_group(stream)
            stream = streams.next()
            next_stamp_and_apid(stream, last_packet.stamp, last_packet.apid)

        except PacketTooShort as err:
            LOG.error("corrupt stream, removing: %s", err)
            streams.invalid(stream)
            stream = streams.next()
            next_stamp_and_apid(stream, last_packet.stamp, last_packet.apid)

        except StopIteration:
            LOG.debug("end-of-stream %s", stream)
            streams.end_of_stream(stream)
            if streams.have_valid():
                stream = streams.next()
                next_stamp(stream, last_packet.stamp)


def merge_files(filepaths, destpath):
    filepaths, streams = make_streams(filepaths, fail_on_missing=True)
    for i in range(0, len(filepaths)):
        LOG.debug("stream %s filepath %s", streams[i], filepaths[i])
    merge(streams, output=io.open(destpath, 'wb'))


def _cmp_by_first_pkt(filea, fileb):
    """
    Compare 2 L0 files by their first available timestamp. Used to sort L0 files
    by time.
    """
    stampa = stampb = None
    for pkt in jpss_packet_stream(io.open(filea, 'rb')):
        if pkt.stamp:
            stampa = pkt.stamp
            break
    for pkt in jpss_packet_stream(io.open(fileb, 'rb')):
        if pkt.stamp:
            stampb = pkt.stamp
            break
    return cmp(stampa, stampb)


def make_streams(*filepaths):
    return [
        jpss_packet_stream(io.open(f, 'rb'))
        for f in sorted(filepaths, cmp=_cmp_by_first_pkt)
    ]
